/* FFT algorithm - from the book DSP by Proakis & Manolakis
 * radix 2 stuff
 * public domain
 * surely slower than death, who gives a shit...
 * sed@free.fr
 */

#include "fft.h"

#ifdef FFTW

#include <stdlib.h>

void fft_memalloc(float **in, fftwf_complex **out, int size)
{
  *in = fftwf_malloc(size * sizeof(float));
  *out = fftwf_malloc((size/2+1) * sizeof(fftwf_complex));
  if (!*in || !*out) abort();
}

void fft_memfree(float *in, fftwf_complex *out)
{
  fftwf_free(in);
  fftwf_free(out);
}

int open_fft(fft *f, int len, float *in, fftwf_complex *out)
{
  f->n = len;
  f->plan = fftwf_plan_dft_r2c_1d(len, in, out, FFTW_ESTIMATE);
  return 0;
}

void close_fft(fft *f)
{
  fftwf_destroy_plan(f->plan);
}

void fft_double(fft *f, float *in, fftwf_complex *out)
{
  fftwf_execute_dft_r2c(f->plan, in, out);
}

#else /* FFTW */

#include <stdlib.h>
#include <math.h>

static unsigned int bit_reverse(unsigned int i, int logn)
{
  unsigned int ret = i;

  ret = ((ret << 1) & 0xAAAAAAAA) | ((ret >> 1) & 0x55555555);
  ret = ((ret << 2) & 0xCCCCCCCC) | ((ret >> 2) & 0x33333333);
  ret = ((ret << 4) & 0xF0F0F0F0) | ((ret >> 4) & 0x0F0F0F0F);
  ret = ((ret << 8) & 0xFF00FF00) | ((ret >> 8) & 0x00FF00FF);
  ret = ((ret << 16) & 0xFFFF0000) | ((ret >> 16) & 0x0000FFFF);

  ret >>= 32-logn;

  return ret;
}

/* return -1 error, 0 ok, alloc some memory with malloc */
int open_fft(fft *f, int n)
{
  int i;
  int logn = 0;
  int nn = n-1;

  /* not really log(n), rather nb of bits of n-1 */
  while (nn) { nn >>= 1; logn++; }

  f->n = n;
  f->logn = logn;

  f->wn = NULL;
  f->bit_reverse = NULL;

  f->wn = malloc(sizeof(complex) * n / 2);
  f->bit_reverse = malloc(sizeof(int) * n);
  if (!f->wn || !f->bit_reverse) goto malloc_error;

  for (i = 0; i < n / 2; i++) {
    f->wn[i].re = cos(M_PI * 2 * i / n);
    f->wn[i].im = sin(M_PI * 2 * i / n);
  }

  for (i = 0; i < n; i++) {
    f->bit_reverse[i] = bit_reverse(i, logn);
  }

  return 0;

malloc_error:
  if (f->wn) free(f->wn);
  if (f->bit_reverse) free(f->bit_reverse);
  return -1;
}

void close_fft(fft *f)
{
  free(f->wn);
  free(f->bit_reverse);
}

/* get a book on FFT, I won't explain stuff here */
static void butterfly(fft *f, complex *v, int a, int b, int k)
{
  complex wn_b;
  complex A, B;

  wn_b.re = f->wn[k].re * v[b].re - f->wn[k].im * v[b].im;
  wn_b.im = f->wn[k].re * v[b].im + f->wn[k].im * v[b].re;

  A.re = v[a].re + wn_b.re;
  A.im = v[a].im + wn_b.im;

  B.re = v[a].re - wn_b.re;
  B.im = v[a].im - wn_b.im;

  v[a] = A;
  v[b] = B;
}

/* fft and ifft are the same except you use W(-k,n) instead
 * of W(k,n), thus the difference in the butterfly calculation
 */
static void ifft_butterfly(fft *f, complex *v, int a, int b, int k)
{
  complex wn_b;
  complex A, B;

  wn_b.re = f->wn[k].re * v[b].re + f->wn[k].im * v[b].im;
  wn_b.im = f->wn[k].re * v[b].im - f->wn[k].im * v[b].re;

  A.re = v[a].re + wn_b.re;
  A.im = v[a].im + wn_b.im;

  B.re = v[a].re - wn_b.re;
  B.im = v[a].im - wn_b.im;

  v[a] = A;
  v[b] = B;
}

void fft_double(fft  *f, double *in, complex *out)
{
  int i, j, k;
  int jinc;
  int kstart;
  int kinc;

  for (i = 0; i < f->n; i++) {
    out[i].re = in[f->bit_reverse[i]];
    out[i].im = 0;
  }

  for (i=0, jinc=f->n/2, kinc=2; i < f->logn; i++, jinc >>= 1, kinc <<= 1)
    for (j = 0, kstart = 0; j < f->n/2; j += jinc, kstart++)
      for (k = kstart; k < f->n; k += kinc)
        butterfly(f, out, k, k + kinc/2, j);
}

void ifft_double(fft  *f, complex *in, complex *out)
{
  int i, j, k;
  int jinc;
  int kstart;
  int kinc;

  for (i = 0; i < f->n; i++) {
    out[i].re = in[f->bit_reverse[i]].re;
    out[i].im = in[f->bit_reverse[i]].im;
  }

  for (i=0, jinc=f->n/2, kinc=2; i < f->logn; i++, jinc >>= 1, kinc <<= 1)
    for (j = 0, kstart = 0; j < f->n/2; j += jinc, kstart++)
      for (k = kstart; k < f->n; k += kinc)
        ifft_butterfly(f, out, k, k + kinc/2, j);

  for (i = 0; i < f->n; i++) {
    out[i].re /= (double)f->n;
    out[i].im /= (double)f->n;
  }
}

#if 0
int main(void)
{
  fft f;
  double x[8];
  complex y[8];
  complex z[8];
  int i;

  x[0] = 1; for (i = 1; i < 8; i++) x[i] = 0;

  open_fft(&f, 8);
  fft_double(&f, x, y);
  ifft_double(&f, y, z);
  close_fft(&f);

  for (i = 0; i < 8; i++) printf("%g ", sqrt(z[i].re*z[i].re+z[i].im*z[i].im));
  printf("\n");

if (1){
#define N 32
  double t[N];
  complex r[N];
  complex s[N];
  for (i=0; i < N; i++) t[i] = 0;
  t[0] = t[14] = -0.014112893;
  t[1] = t[13] = -0.001945309;
  t[2] = t[12] = 0.04000004;
  t[3] = t[11] = 0.01223454;
  t[4] = t[10] = -0.09138802;
  t[5] = t[9] = -0.01808986;
  t[6] = t[8] = 0.3133176;
  t[7] = 0.52;
  open_fft(&f, N);
  fft_double(&f, t, r);
  ifft_double(&f, r, s);
  close_fft(&f);
  for (i=0; i<N/2; i++)
    printf("%d %g\n", i, 20 * log10(sqrt(r[i].re*r[i].re+r[i].im*r[i].im)));
  for (i=0; i<(15>N?N:15); i++)
    printf("%d %g\n", i, s[i].re);
}

  return 0;
}
#endif

#endif /* FFTW */
