#include "power_buffer.h"

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "fft.h"

void power_buffer_init(power_buffer *b, int width, int fft_size,
                     void (*bnl)(void *), void *bnl_data, int window_type)
{
  int i;
  b->buffer_new_line = bnl;
  b->bnl_data = bnl_data;
  b->width = width;
  b->power = malloc(sizeof(float) * width * (fft_size/2+1));
#ifdef FFTW
  fft_memalloc(&b->buffer, &b->fft_buffer, fft_size);
#else
  b->buffer = malloc(sizeof(fft_in) * fft_size);
  b->fft_buffer = malloc(sizeof(complex) * fft_size);
#endif
  if (!b->power || !b->buffer || !b->fft_buffer) { printf("oom\n"); exit(1); }
  for (i = 0; i < fft_size; i++) b->buffer[i] = 0;
  for (i = 0; i < width * (fft_size/2+1); i++) b->power[i] = -120;
  b->processed = 0;
  b->to_process = fft_size;
  b->overlap = 0;
  b->overlap_samples = 0;
#ifdef FFTW
  if (open_fft(&b->fft, fft_size, b->buffer, b->fft_buffer))
    { printf("err open fft\n"); exit(1); }
#else
  if (open_fft(&b->fft, fft_size)) { printf("err open fft\n"); exit(1); }
#endif
  b->_10_log10_nn = 20 * log10f(fft_size);
  b->line = 0;
  b->buf_min = 10000;
  b->buf_max = -10000;
  b->mindB_stats = 10000;
  b->maxdB_stats = -10000;
  fft_window_init(&b->window, b->to_process, window_type);
}

static void compute_power(power_buffer *b)
{
  float *buf = b->power + b->line * (b->fft.n/2+1);
  int i;
  fft_window_mul(b->buffer, &b->window);
  fft_double(&b->fft, b->buffer, b->fft_buffer);
  for (i = 0; i < b->fft.n/2+1; i++) {
    buf[i] =
#ifdef FFTW
        10 * log10f(1e-40+(b->fft_buffer[i][0] * b->fft_buffer[i][0] +
         b->fft_buffer[i][1] * b->fft_buffer[i][1])) - b->_10_log10_nn;
#else
        10 * log10f(1e-40+(b->fft_buffer[i].re * b->fft_buffer[i].re +
         b->fft_buffer[i].im * b->fft_buffer[i].im)) - b->_10_log10_nn;
#endif
    if (buf[i] < b->mindB_stats) b->mindB_stats = buf[i];
    if (buf[i] > b->maxdB_stats) b->maxdB_stats = buf[i];
  }
  b->line++;
  if (b->line == b->width) b->line = 0;
}

void power_buffer_feed(power_buffer *b, float *buf, int len)
{
  int i;
  for (i = 0; i < len; i++) {
    if (b->processed == b->to_process) {
      b->processed = b->overlap_samples;
      compute_power(b);
      b->buffer_new_line(b->bnl_data);
      memmove(b->buffer, b->buffer + b->to_process - b->overlap_samples,
              b->overlap_samples * sizeof(fft_in));
    }
    b->buffer[b->processed] = buf[i];
    if (buf[i] < b->buf_min) b->buf_min = buf[i];
    if (buf[i] > b->buf_max) b->buf_max = buf[i];
    b->processed++;
  }
}

float *power_buffer_get_line(power_buffer *b, int l)
{
  l += b->line;
  if (l >= b->width) l -= b->width;
  return b->power + l * (b->fft.n/2+1);
}

void power_buffer_set_to_process(power_buffer *b, int s)
{
  int i;
  if (s < 128) s = 128;
  if (s > b->fft.n) s = b->fft.n;
  b->to_process = s;
  if (b->processed > b->to_process) b->processed = b->to_process;
  for (i = b->to_process; i < b->fft.n; i++) b->buffer[i] = 0;
  power_buffer_set_overlap(b, b->overlap);
  fft_window_close(&b->window);
  fft_window_init(&b->window, b->to_process, b->window.type);
}

void power_buffer_set_overlap(power_buffer *b, float o)
{
  if (o < 0) o = 0;
  if (o > 1) o = 1;
  b->overlap_samples = b->to_process * o;
  if (b->overlap_samples == b->to_process)
    b->overlap_samples = b->to_process - 1;
  b->overlap = o;
}

void power_buffer_close(power_buffer *b)
{
  free(b->power);
#ifdef FFTW
  fft_memfree(b->buffer, b->fft_buffer);
#else
  free(b->buffer);
  free(b->fft_buffer);
#endif
  close_fft(&b->fft);
  fft_window_close(&b->window);
}
